package autogroup;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.ArrayList;
//import java.util.Collections;
//import java.util.List;
import java.util.Random;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import autogroup.Groupe;
import autogroup.Eleve;

public class Magic {
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {

		//CommandLineParser parser = new PosixParser();

		Option help = new Option( "help", "Affiche ce message" );

		@SuppressWarnings("static-access")
		Option inputfile = OptionBuilder.withArgName( "list" )
		.hasArg()
		.withDescription(  "Csv file containing the list of students (default: list.csv)" )
		.create( "inputfile" );

		@SuppressWarnings("static-access")
		Option outputfile = OptionBuilder.withArgName( "out" )
		.hasArg()
		.withDescription( "Csv output file with the resulting groups (default: groups.csv)" )
		.create( "outputfile" );

		@SuppressWarnings("static-access")
		Option paramfile  = OptionBuilder.withArgName( "params" )
		.hasArg()
		.withDescription( "Settings file (default: params.txt)" )
		.create( "paramfile"); 

		Option quiet = new Option( "quiet", "Be extra quiet" );
		Option verbose = new Option( "verbose", "Be extra verbose" );
		
		Option reverse = new Option("reverse", "Make homogen groups instead");

		Options options = new Options();

		options.addOption( help );
		options.addOption( inputfile );
		options.addOption( outputfile );
		options.addOption( paramfile );
		options.addOption( reverse );
		options.addOption( quiet );
		options.addOption( verbose );

		// create the parser
		CommandLineParser parser = new GnuParser();
		try {
			// parse the command line arguments
			CommandLine line = parser.parse( options, args );
			// has the help argument been passed?
			if( line.hasOption( "help" ) ) {
				HelpFormatter formatter = new HelpFormatter();
				formatter.printHelp( "AutoGroup", options );
			} else {


				ArrayList<Groupe> l = new ArrayList<Groupe>();
				ArrayList<Groupe> lb = new ArrayList<Groupe>();

				int n = 3; // Nombre de valeurs possibles pour les critres
				
				if ( line.hasOption( "reverse" ) ) {
					// initialise inputfile variable
					n = -n;
				}	    
				
				double score = 0.0;
				double bestscore = 0.0;
				//int iter = 0;
				int numEleves = 0;
				int nbGroupes = 0;

				Random randomGenerator = new Random();

				int a=0;
				int b=0;

				// Lire les donnes

				String fileName = "list.csv";

				// S'il y a un fichier donn en entre
				if( line.hasOption( "inputfile" ) ) {
					// initialise inputfile variable
					fileName = line.getOptionValue( "inputfile" );
				}	    		

				CSVFileReader x=new CSVFileReader(fileName);
				x.ReadFile();
				//x.displayArrayList();
				numEleves = x.getNumLines();
				@SuppressWarnings("unchecked")
				ArrayList<String> ar = x.getFileValues();

				// Lire les parametres (nbgroupes, taille des groupes)

				String fileParams="params.txt";
				if( line.hasOption( "paramfile" ) ) {
					// initialise outputfile variable
					fileParams = line.getOptionValue( "paramfile" );
				}

				CSVFileReader par=new CSVFileReader(fileParams);
				par.ReadFile();
				@SuppressWarnings("unchecked")
				ArrayList<String> params = par.getFileValues();

				nbGroupes = Integer.parseInt(params.get(0));

				int nbMaxEleves = 0;
				for (int nb=0;nb<nbGroupes;nb++){
					nbMaxEleves += Integer.parseInt(params.get(nb+1));
				}

				if (nbMaxEleves == numEleves) {

					int nel = 0;
					for (int nb=0;nb<nbGroupes;nb++){
						l.add(new Groupe());
						for (int ne=0;ne<Integer.parseInt(params.get(nb+1));ne++){
							l.get(nb).putEleve(new Eleve(ar.get(nel)));
							nel++;
						}
					}

					// Gnrer une partition initiale

					a = randomGenerator.nextInt(l.size());
					b=a;


					// Calculer le score initial de la partition


					for (int i=0;i<l.size();i++){
						score += l.get(i).score(n);
					}
					bestscore = score;
					
					int kmax = 500;
					double T = Math.abs(score)*10+100;
					double lambda = 0.95;
					
					int tdt = (int) Math.ceil(Math.log(1/T)/Math.log(lambda));
					int td = (int) Math.ceil(Math.log(1/T)/Math.log(lambda));
					td = (int) Math.ceil(100*td/tdt);
					
					while(T>1){ //boucle exterieure du recuit
						for (int k=0;k<kmax;k++){
							a = randomGenerator.nextInt(l.size());
							while(a==b) { // On s'assure que a et b sont diffrents
								b = randomGenerator.nextInt(l.size());
							}
							ArrayList<Groupe> l0 = new ArrayList<Groupe>();
							deepClone(l0, l);
							permuter(l0.get(a),l0.get(b));
							double newscore = calcScore(l0,n);
							double oldscore = calcScore(l,n);
							if (accepter(oldscore-newscore,T)) {
								if (newscore > bestscore) {
									bestscore = newscore;
									deepClone(lb, l0);
								}
								deepClone(l, l0);
							}
						}
						T = lambda*T;
						td = (int) Math.ceil(Math.log(1/T)/Math.log(lambda));
						td = (int) Math.ceil(100*td/tdt);
						if( line.hasOption( "verbose" ) &&  !line.hasOption( "quiet" )){
							
							System.out.print("\r Meilleur score : " + (int)bestscore + " [" + (100-td) + "%]   ");
						} else {
							System.out.print("\r[" + (100-td) + "%]   ");
						}
						//System.out.println("Temprature : "+ T);
					}

					score = calcScore(l,n);

					try{
						String outFile = "groups.csv";
						if( line.hasOption( "outputfile" ) ) {
							// initialise outputfile variable
							outFile = line.getOptionValue( "outputfile" );
						}
						// Create file 

						FileWriter fstream = new FileWriter(outFile);
						BufferedWriter out = new BufferedWriter(fstream);
						out.write("Meilleur score " + score + System.getProperty("line.separator"));
						for (int i=0;i<lb.size();i++){
							out.write("Groupe "+ (i+1) + " :" + System.getProperty("line.separator"));
							for (int j=0;j<lb.get(i).getTaille();j++){
								out.write(lb.get(i).getEleve(j).getNom() + ";" +
										lb.get(i).getEleve(j).getCrit1() + ";" +
										lb.get(i).getEleve(j).getCrit2() + ";" +
										lb.get(i).getEleve(j).getCrit3() + ";" +
										lb.get(i).getEleve(j).getCrit4() + ";" +
										lb.get(i).getEleve(j).getClasse() + System.getProperty("line.separator"));
							}
						}
						//Close the output stream
						out.close();
					} catch (Exception e){//Catch exception if any
						System.err.println("Error: " + e.getMessage());
					}

					if( !line.hasOption( "quiet" ) ) {						

						System.out.println("\rMeilleur score : " + score + "\n");
						for (int i=0;i<lb.size();i++){
							System.out.println("Groupe "+ (i+1) + " :");
							for (int j=0;j<lb.get(i).getTaille();j++){
								System.out.println(lb.get(i).getEleve(j).getNom() + " (" + lb.get(i).getEleve(j).getClasse() + ")");
							}

						}
					}

				} else {

					System.out.println("Erreur : le nombre de places est diffrent du nombre d'lves");
				}


			}
		}
		catch( ParseException exp ) {
			// oops, something went wrong
			System.err.println( "Parsing failed.  Reason: " + exp.getMessage() );
		}


	}

	static void permuter(Groupe g1, Groupe g2){
		Random randomGenerator = new Random();

		int a = randomGenerator.nextInt(g1.getTaille());
		int b = randomGenerator.nextInt(g2.getTaille());

		Eleve e1 = g1.getEleve(a);
		Eleve e2 = g2.getEleve(b);

		g1.retireEleve(e1);
		g2.retireEleve(e2);

		g1.putEleve(e2);
		g2.putEleve(e1);

	}

	static void deepClone(ArrayList<Groupe> n, ArrayList<Groupe> o){
		int size = o.size();

		n.clear();

		for (int i=0;i<size;i++){
			n.add(new Groupe());
			for (int j=0;j<o.get(i).getTaille();j++){
				Eleve e = new Eleve();
				deepClone(e,o.get(i).getEleve(j));
				n.get(i).putEleve(e);
			}

		}

	}

	static void deepClone(Eleve n, Eleve o){
		n.setNom(o.getNom());
		n.setCrit1(o.getCrit1());
		n.setCrit2(o.getCrit2());
		n.setCrit3(o.getCrit3());
		n.setCrit4(o.getCrit4());
		n.setClasse(o.getClasse());
	}

	static double calcScore(ArrayList<Groupe> l, int n){
		double score=0;
		for (int i=0;i<l.size();i++){
			score += l.get(i).score(n);
		}
		return score;
	}

	static boolean accepter(double dt, double T){
		//System.out.println("Proba : " + Math.exp(-dt/T));
		if (dt < 0) return true;
		double A = Math.exp(-dt/T);
		Random randomGenerator = new Random();
		double alea = randomGenerator.nextDouble();
		if (alea<A) return true;
		return false;
	}
}
